@@ -1,14 +1,22 @@
 from math import ceil
-from functools import partial
 import torch
 from torch import nn
 import torch.nn.functional as F
-from torch.nn.utils.rnn import pad_sequence
-from entmax import entmax_bisect
+
+from einops import rearrange, pack, unpack
 
 def exists(val):
     return val is not None
 
+def eval_decorator(fn):
+    def inner(self, *args, **kwargs):
+        was_training = self.training
+        self.eval()
+        out = fn(self, *args, **kwargs)
+        self.train(was_training)
+        return out
+    return inner
+
 # nucleus
 
 def top_p(logits, thres = 0.9):
@@ -36,17 +44,20 @@ def top_k(logits, thres = 0.9):
 def top_a(logits, min_p_pow=2.0, min_p_ratio=0.02):
     probs = F.softmax(logits, dim=-1)
     limit = torch.pow(torch.max(probs), min_p_pow) * min_p_ratio
-    logits[probs < limit] = -float("Inf")
+    logits[probs < limit] = float('-inf')
     logits[probs >= limit] = 1
     return logits
 
-# entmax
-
-ENTMAX_ALPHA = 1.3
-entmax = entmax_bisect
+# autoregressive wrapper class
 
 class AutoregressiveWrapper(nn.Module):
-    def __init__(self, net, ignore_index = -100, pad_value = 0):
+    def __init__(
+        self,
+        net,
+        ignore_index = -100,
+        pad_value = 0,
+        mask_prob = 0.
+    ):
         super().__init__()
         self.pad_value = pad_value
         self.ignore_index = ignore_index
@@ -54,29 +65,37 @@ class AutoregressiveWrapper(nn.Module):
         self.net = net
         self.max_seq_len = net.max_seq_len
 
-    @torch.no_grad()
-    def generate(self, start_tokens, seq_len, eos_token = None, temperature = 1., filter_logits_fn = top_k, filter_thres = 0.9, min_p_pow=2.0, min_p_ratio=0.02, **kwargs):
-        device = start_tokens.device
-        was_training = self.net.training
-        num_dims = len(start_tokens.shape)
+        # paper shows masking (MLM) in conjunction with autoregressive decoder-only training leads to big improvements https://arxiv.org/abs/2210.13432
+        assert mask_prob < 1.
+        self.mask_prob = mask_prob
 
-        if num_dims == 1:
-            start_tokens = start_tokens[None, :]
+    @torch.no_grad()
+    @eval_decorator
+    def generate(
+        self,
+        start_tokens,
+        seq_len,
+        eos_token = None,
+        temperature = 1.,
+        filter_logits_fn = top_k,
+        filter_thres = 0.9,
+        min_p_pow = 2.0,
+        min_p_ratio = 0.02,
+        **kwargs
+    ):
+        device = start_tokens.device
+        num_dims = start_tokens.ndim
+
+        start_tokens, ps = pack([start_tokens], '* n')
 
         b, t = start_tokens.shape
 
-        self.net.eval()
         out = start_tokens
-        mask = kwargs.pop('mask', None)
-
-        if mask is None:
-            mask = torch.full_like(out, True, dtype=torch.bool, device=out.device)
 
         for _ in range(seq_len):
             x = out[:, -self.max_seq_len:]
-            mask = mask[:, -self.max_seq_len:]
 
-            logits = self.net(x, mask=mask, **kwargs)[:, -1, :]
+            logits = self.net(x, **kwargs)[:, -1]
 
             if filter_logits_fn in {top_k, top_p}:
                 filtered_logits = filter_logits_fn(logits, thres = filter_thres)
@@ -86,13 +105,9 @@ class AutoregressiveWrapper(nn.Module):
                 filtered_logits = filter_logits_fn(logits, min_p_pow = min_p_pow, min_p_ratio= min_p_ratio)
                 probs = F.softmax(filtered_logits / temperature, dim=-1)
 
-            elif filter_logits_fn is entmax:
-                probs = entmax(logits / temperature, alpha = ENTMAX_ALPHA, dim=-1)
-
             sample = torch.multinomial(probs, 1)
 
             out = torch.cat((out, sample), dim=-1)
-            mask = F.pad(mask, (0, 1), value=True)
 
             if exists(eos_token):
                 is_eos_tokens = (out == eos_token)
@@ -106,23 +121,29 @@ class AutoregressiveWrapper(nn.Module):
 
         out = out[:, t:]
 
-        if num_dims == 1:
-            out = out.squeeze(0)
+        out, = unpack(out, ps, '* n')
 
-        self.net.train(was_training)
         return out
 
     def forward(self, x, **kwargs):
-        xi = x[:, :-1]
-        xo = x[:, 1:]
+        seq, ignore_index = x.shape[1], self.ignore_index
 
-        # help auto-solve a frequent area of confusion around input masks in auto-regressive
-        # if user supplies a mask that is only off by one from the source sequence, resolve it for them
-        mask = kwargs.get('mask', None)
-        if mask is not None and mask.shape[1] == x.shape[1]:
-            mask = mask[:, :-1]
-            kwargs['mask'] = mask
+        inp, target = x[:, :-1], x[:, 1:]
+
+        if self.mask_prob > 0.:
+            rand = torch.randn(inp.shape, device = x.device)
+            rand[:, 0] = -torch.finfo(rand.dtype).max # first token should not be masked out
+            num_mask = min(int(seq * self.mask_prob), seq - 1)
+            indices = rand.topk(num_mask, dim = -1).indices
+            mask = ~torch.zeros_like(inp).scatter(1, indices, 1.).bool()
+            kwargs.update(self_attn_context_mask = mask)
+
+        logits = self.net(inp, **kwargs)
+
+        loss = F.cross_entropy(
+            rearrange(logits, 'b n c -> b c n'),
+            target,
+            ignore_index = ignore_index
+        )
 
-        out = self.net(xi, **kwargs)
-        loss = F.cross_entropy(out.transpose(1, 2), xo, ignore_index = self.ignore_index)
         return loss
